from __future__ import annotations

import math
from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING

import optiland.backend
import pytest

from tests.helpers import build_args

if TYPE_CHECKING:
    from optiland.fields import FieldGroup

    from visisipy import EyeModel
    from visisipy.optiland.backend import OptilandBackend, OptilandSettings


class TestOptilandBackend:
    def test_initialize_optiland(self, optiland_backend: OptilandBackend):
        assert optiland_backend.optic is not None
        assert optiland_backend.model is None

    def test_new_model(self, optiland_backend: OptilandBackend):
        # Change a setting and add a new surface
        optiland_backend.get_optic().add_field(10, 10)
        optiland_backend.get_optic().add_surface(index=0)

        assert optiland_backend.get_optic().surface_group.num_surfaces == 1

        optiland_backend.new_model(save_old_model=False)

        assert optiland_backend.get_optic().surface_group.num_surfaces == 0
        assert optiland_backend.get_optic().fields.num_fields == 1

    def test_build_model(self, optiland_backend: OptilandBackend, eye_model: EyeModel):
        optiland_backend.build_model(eye_model)

        assert optiland_backend.model is not None
        assert optiland_backend.model.eye_model == eye_model
        assert optiland_backend.get_optic().surface_group.num_surfaces == 7

    def test_clear_model(self, optiland_backend: OptilandBackend, eye_model: EyeModel):
        optiland_backend.build_model(eye_model)
        assert optiland_backend.model is not None
        assert optiland_backend.get_optic().surface_group.num_surfaces == 7

        optiland_backend.clear_model()

        assert optiland_backend.model is None
        assert optiland_backend.get_optic().surface_group.num_surfaces == 0

    def test_save_model(self, optiland_backend: OptilandBackend, eye_model, tmp_path):
        optiland_backend.build_model(eye_model)

        path = tmp_path / "test_model.json"
        optiland_backend.save_model(path=path)

        assert path.exists()

    def test_save_model_invalid_filename_raises_valueerror(self, optiland_backend: OptilandBackend):
        with pytest.raises(ValueError, match=r"filename must end in \.json"):
            optiland_backend.save_model(path="invalid_path/test_model.zmx")

    @pytest.mark.parametrize(
        "coordinates,field_type,expectation",
        [
            ([(10, 10)], "angle", does_not_raise()),
            (
                [(0, 0), (0, 10), (-10, 0), (10, -10)],
                "angle",
                does_not_raise(),
            ),
            (
                [(math.pi, 4.15), (-12, 1)],
                "object_height",
                does_not_raise(),
            ),
            (
                [(0, 0)],
                "invalid",
                pytest.raises(
                    ValueError,
                    match="field_type must be either 'angle' or 'object_height'",
                ),
            ),
        ],
    )
    def test_set_fields(self, coordinates, field_type, expectation, optiland_backend: OptilandBackend):
        with expectation:
            optiland_backend.set_fields(coordinates, field_type)

            assert optiland_backend.get_optic().fields.num_fields == len(coordinates)
            for coordinate, field in zip(coordinates, optiland_backend.get_optic().fields.fields):
                assert coordinate == (field.x, field.y)

            assert all(f.field_type == field_type for f in optiland_backend.get_optic().fields.fields)

    def test_get_fields(self, optiland_backend: OptilandBackend):
        coordinates = [(10, 10), (20, 20)]
        field_type = "angle"

        optiland_backend.get_optic().fields.fields.clear()
        optiland_backend.get_optic().set_field_type(field_type)

        for coordinate in coordinates:
            optiland_backend.get_optic().add_field(coordinate[0], coordinate[1])

        assert optiland_backend.get_fields() == coordinates

    def test_set_wavelengths(self, optiland_backend: OptilandBackend):
        wavelengths = [0.543, 0.650, 0.450]

        optiland_backend.set_wavelengths(wavelengths)

        assert optiland_backend.get_optic().wavelengths.num_wavelengths == len(wavelengths)
        assert all(w.value == e for w, e in zip(optiland_backend.get_optic().wavelengths.wavelengths, wavelengths))

    def test_get_wavelengths(self, optiland_backend: OptilandBackend):
        wavelengths = [0.543, 0.650, 0.450]

        optiland_backend.get_optic().wavelengths.wavelengths.clear()

        for wavelength in wavelengths:
            optiland_backend.get_optic().add_wavelength(wavelength)

        assert optiland_backend.get_wavelengths() == wavelengths


class TestOptilandBackendSettings:
    def test_update_settings(self, optiland_backend):
        settings: OptilandSettings = {
            "field_type": "object_height",
            "fields": [(0, 0), (0, 10), (-10, 0), (10, -10)],
            "wavelengths": [0.55, 0.65, 0.75],
            "aperture_type": "entrance_pupil_diameter",
            "aperture_value": 3.0,
            "computation_backend": "numpy",
        }

        optiland_backend.update_settings(**settings)

        assert all(optiland_backend.settings[k] == v for k, v in settings.items())

    @staticmethod
    def _assert_fields_equal(optic_fields: FieldGroup, expected_fields: list[tuple[float, float]]):
        assert len(optic_fields.fields) == len(expected_fields)

        for optiland_field, expected_field in zip(optic_fields.fields, expected_fields):
            assert optiland_field.x == expected_field[0]
            assert optiland_field.y == expected_field[1]

    @pytest.mark.parametrize(
        "field_type,fields",
        [
            ("angle", [(10, 10)]),
            ("object_height", [(0, 0), (0, 10), (-10, 0), (10, -10)]),
        ],
    )
    def test_field(self, field_type, fields, optiland_backend: OptilandBackend):
        field_type = "angle"
        fields = [(10, 10), (20, 20)]

        optiland_backend.update_settings(field_type=field_type, fields=fields)

        self._assert_fields_equal(optiland_backend.get_optic().fields, fields)
        assert all(f.field_type == field_type for f in optiland_backend.get_optic().fields.fields)

    @pytest.mark.parametrize(
        "wavelengths",
        [[0.543, 0.650], [0.543, 0.650, 0.450]],
    )
    def test_wavelength(self, wavelengths, optiland_backend: OptilandBackend):
        optiland_backend.update_settings(wavelengths=wavelengths)

        assert all(w.value == e for w, e in zip(optiland_backend.get_optic().wavelengths.wavelengths, wavelengths))

    @pytest.mark.parametrize(
        "aperture_type,aperture_value,expected_aperture_type,expectation",
        [
            ("entrance_pupil_diameter", 3.0, "EPD", does_not_raise()),
            ("image_f_number", 1.0, "imageFNO", does_not_raise()),
            ("object_numeric_aperture", 0.1, "objectNA", does_not_raise()),
            (
                "float_by_stop_size",
                None,
                "FloatByStopSize",
                pytest.raises(
                    NotImplementedError, match="Aperture type 'float_by_stop_size' is not implemented in Optiland"
                ),
            ),
            (
                "invalid_aperture_type",
                3.0,
                "InvalidApertureType",
                pytest.raises(ValueError, match="Invalid aperture type"),
            ),
        ],
    )
    def test_aperture(
        self,
        aperture_type,
        aperture_value,
        expected_aperture_type,
        expectation,
        optiland_backend: OptilandBackend,
        monkeypatch,
    ):
        with expectation:
            optiland_backend.update_settings(aperture_type=aperture_type, aperture_value=aperture_value)

            assert optiland_backend.get_optic().aperture.ap_type == expected_aperture_type
            assert optiland_backend.get_optic().aperture.value == aperture_value

    @pytest.mark.parametrize(
        "backend,torch_device,torch_precision,torch_gradient_mode,expectation",
        [
            ("numpy", None, None, None, does_not_raise()),
            ("torch", "cpu", "float64", False, does_not_raise()),
            ("torch", "cuda", "float32", True, does_not_raise()),
            (
                "invalid_backend",
                "cpu",
                "float32",
                "none",
                pytest.raises(ValueError, match="computation_backend must be either 'numpy' or 'torch'"),
            ),
            (
                "torch",
                "invalid_device",
                "float32",
                True,
                pytest.raises(ValueError, match="torch_device must be either 'cpu' or 'cuda'"),
            ),
            (
                "torch",
                "cpu",
                "invalid_precision",
                True,
                pytest.raises(ValueError, match="torch_precision must be either 'float32' or 'float64'"),
            ),
        ],
    )
    def test_computation_backend(
        self, backend, torch_device, torch_precision, torch_gradient_mode, expectation, optiland_backend, monkeypatch
    ):
        if backend == "torch":
            torch = pytest.importorskip("torch", reason="Torch is not installed")
            monkeypatch.setattr(torch.cuda, "is_available", lambda: True)

        args = build_args(
            computation_backend=backend,
            torch_device=torch_device,
            torch_precision=torch_precision,
            torch_use_grad_mode=torch_gradient_mode,
            non_null_defaults={"computation_backend", "torch_device", "torch_precision", "torch_use_grad_mode"},
        )

        with expectation:
            optiland_backend.update_settings(**args)

            assert optiland.backend.get_backend() == backend

            if torch_device is not None:
                assert optiland.backend.get_device() == torch_device
            if torch_precision is not None:
                assert str(optiland.backend.get_precision()) == f"torch.{torch_precision}"
            if torch_gradient_mode is not None:
                assert optiland.backend.grad_mode.requires_grad == torch_gradient_mode



# %%
from __future__ import annotations

from operator import itemgetter
from typing import NamedTuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import visisipy

sns.set_style(
    "whitegrid",
    rc={
        "axes.grid": True,
        "grid.linestyle": ":",
        "grid.color": "gray",
        "grid.alpha": 0.5,
        "patch.edgecolor": "black",
    },
)
# %%
# Initialize the default Navarro model
model = visisipy.EyeModel()

# Build the model in OpticStudio
model.build()

# Add dummy surface
dummy_surface = visisipy.get_backend().oss.LDE.InsertNewSurfaceAt(1)
dummy_surface.Comment = "dummy"
dummy_surface.Thickness = 1

# %% Perform a raytrace analysis
y_angles = range(0, 90, 5)
raytrace = visisipy.analysis.raytrace(coordinates=zip([0] * len(y_angles), y_angles))

raytrace["input_angle"] = raytrace.field.apply(itemgetter(1))


# %% Calculate input and output angles
class InputOutputAngles(NamedTuple):
    input_angle: float
    output_angle_np2: float
    output_angle_pupil: float
    output_angle_retina_center: float

    @staticmethod
    def output_angle(coordinate: tuple[float, float], reference_coordinate: tuple[float, float]) -> float:
        return np.rad2deg(
            np.arctan2(
                coordinate[1] - reference_coordinate[1],
                coordinate[0] - reference_coordinate[0],
            )
        )

    @classmethod
    def from_raytrace(
        cls,
        raytrace: pd.DataFrame,
        location_np2: float,
        location_pupil: float,
        location_retina_center: float,
        coordinate: str = "y",
    ):
        image_coordinate: tuple[float, float] = tuple(raytrace.query("comment == 'retina'").iloc[0][["z", coordinate]])

        return cls(
            input_angle=raytrace.field.iloc[0]["xy".index(coordinate)],
            output_angle_np2=cls.output_angle(image_coordinate, (location_np2, 0)),
            output_angle_pupil=cls.output_angle(image_coordinate, (location_pupil, 0)),
            output_angle_retina_center=cls.output_angle(image_coordinate, (location_retina_center, 0)),
        )


location_np2 = 7.45 - (model.geometry.cornea_thickness + model.geometry.anterior_chamber_depth)
location_retina_center = (
    model.geometry.lens_thickness + model.geometry.vitreous_thickness - abs(model.geometry.retina.ellipsoid_radii.z)
)

angles = pd.DataFrame(
    InputOutputAngles.from_raytrace(rt, location_np2, 0, location_retina_center)
    for _, rt in raytrace.groupby("input_angle")
)

# %% Visualize the model and angle relations
fig, ax = plt.subplots(nrows=1, ncols=2, figsize=(8, 4), gridspec_kw={"width_ratios": [1, 0.8]})

sns.lineplot(
    raytrace,
    x="z",
    y="y",
    hue="input_angle",
    legend="brief",
    palette="plasma",
    ax=ax[0],
)
sns.move_legend(ax[0], "lower right", title="input angle [°]")

visisipy.plots.plot_eye(ax[0], model.geometry, lens_edge_thickness=0.5, zorder=5)
ax[0].set_xlim((-7, 23))
ax[0].set_ylim((-15, 15))
ax[0].set_aspect("equal")
ax[0].set_xlabel("z [mm]")
ax[0].set_ylabel("y [mm]")

sns.lineplot(
    data=angles,
    x="input_angle",
    y="output_angle_np2",
    label="second nodal point",
    ax=ax[1],
)
sns.lineplot(
    data=angles,
    x="input_angle",
    y="output_angle_pupil",
    label="pupil",
    ax=ax[1],
)
sns.lineplot(
    data=angles,
    x="input_angle",
    y="output_angle_retina_center",
    label="retina center",
    ax=ax[1],
)
ax[1].set_aspect("equal")
ax[1].set_xlabel("input angle [°]")
ax[1].set_ylabel("output angle [°]")

fig.savefig("visisipy_example.png", dpi=600, bbox_inches="tight")

plt.show()

# %%



"""Visualization tools for Visisipy."""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import numpy as np
from matplotlib import patches
from matplotlib.path import Path
from scipy.optimize import fsolve

from visisipy.models import EyeGeometry, EyeModel

if TYPE_CHECKING:
    from matplotlib.axes import Axes


__all__ = ("plot_eye",)


def plot_surface(
    position: float,
    radius: float,
    conic: float,
    cutoff: float,
    *,
    return_endpoint: bool = False,
) -> Path | tuple[Path, float]:
    """Plot a conic surface.

    Creates a MatPlotLib `Path` for a conic segment, intersecting the horizontal axis in `position`.
    The radius of curvature at the apex is specified as `radius`, the asfericity as `conic`.
    The segment is cut off at the x-coordinate specified by `cutoff`.
    The orientation of the surface is controlled with the sign of `radius`.

    Returns
    -------
    Path
        A `matplotlib.path.Path` object with the surface.

    Arguments
    ---------
    position : float
        Position of the surface at the optical axis.
    radius : float
        Radius of curvature at the intersection with the horizontal axis.
    conic : float
        Conic constant (asphericity) of the surface. The surface is a hyperbola for `conic < -1`, a
        parabola for `conic == -1` and an ellipse for `conic > -1`.
    cutoff : float
        Position coordinate at which the surface is cut off.
    return_endpoint : bool
        If true, returns the y coordinate of the arc endpoint.
    """
    if conic > -1:
        return plot_ellipse(position, radius, conic, cutoff, return_endpoint=return_endpoint)
    if conic == -1:
        return plot_parabola(position, radius, cutoff, return_endpoint=return_endpoint)

    return plot_hyperbola(position, radius, conic, cutoff)


def _get_ellipse_sizes(radius: float, conic: float) -> tuple[float, float]:
    """Calculate rx and ry (axial and radial radii) of an ellipse from its radius of curvature and conic."""
    # Prolate or oblate does not matter in this case
    return radius / (conic + 1), radius / np.sqrt(conic + 1)


def plot_ellipse(
    position: float,
    radius: float,
    conic: float,
    cutoff: float,
    *,
    return_endpoint: bool = False,
) -> Path | tuple[Path, float]:
    """Plot a segment of an ellipse.

    Creates an `Arc` patch for an ellipse. The radius of curvature at the apex is specified as `radius`.
    The ellipse is cut off at the x-coordinate specified by `cutoff`.

    Returns
    -------
    Path
        A `matplotlib.path.Path` object with the ellipse segment.
    float
        y-coordinate of the segment's end point.

    Raises
    ------
    ValueError
        If the cutoff is located outside the ellipse.

    Arguments
    ---------
    position : float
        Coordinate of the ellipse apex.
    radius : float
        Radius of curvature at the apex.
    conic : float
        Conic constant (asphericity) of the ellipse. Must be > -1.
    cutoff : float
        x-coordinate at which the ellipse is cut off.
    return_endpoint : bool
        If true, returns the y coordinate of the arc endpoint.
    """
    rx, ry = _get_ellipse_sizes(radius, conic)

    x0 = position + rx

    if abs(cutoff) > abs(rx):
        message = f"cutoff is located outside the ellipse: {cutoff=}, {rx=}"
        raise ValueError(message)

    t_max = np.abs(np.arccos((cutoff - x0) / rx))
    # This is a bit weird, but necessary to draw the arc in the right direction
    t = np.linspace(-t_max, t_max - 2 * np.pi, 1000)

    vertices = list(zip(x0 + rx * np.cos(t), ry * np.sin(t)))
    codes = [Path.MOVETO] + [Path.LINETO] * (len(vertices) - 1)
    ellipse = Path(vertices, codes)

    return (ellipse, ry * np.sin(t_max)) if return_endpoint else ellipse


def plot_parabola(
    position: float,
    radius: float,
    cutoff: float,
    *,
    return_endpoint: bool = False,  # noqa: ARG001
) -> Path:
    """Plot a segment of a parabola.

    Creates a `Path` for a parabola. The radius of curvature at the apex is specified as `radius`.
    The parabola is cut off at the x-coordinate specified by `cutoff`.

    Returns
    -------
    Path
        A `matplotlib.path.Path` object with the parabola segment.
    float
        y-coordinate of the segment's end point.

    Raises
    ------
    ValueError
        If the cutoff is located outside the parabola.

    Arguments
    ---------
    position : float
        Coordinate of the ellipse apex.
    radius : float
        Radius of curvature at the apex.
    cutoff : float
        x-coordinate at which the ellipse is cut off.
    return_endpoint : bool
        If true, returns a the y coordinate of the arc endpoint.
    """
    a = radius / 2  # The radius of curvature of a parabola is twice its focal length

    if (cutoff < position and radius > 0) or (cutoff > position and radius < 0):
        message = "The cutoff coordinate is located outside the domain of the parabola."
        raise ValueError(message)

    t_max = np.abs(np.sqrt((cutoff - position) / a))
    t = np.linspace(-t_max, t_max, 1000)

    vertices = list(zip(position + a * t**2, 2 * a * t))
    codes = [Path.MOVETO] + [Path.LINETO] * (len(vertices) - 1)

    return Path(vertices, codes)


def _get_hyperbola_sizes(radius: float, conic: float) -> tuple[float, float]:
    return -radius / (conic + 1), radius / np.sqrt(-conic - 1)


def plot_hyperbola(position: float, radius: float, conic: float, cutoff: float) -> Path:
    """Plot a segment of a hyperbola.

    Creates a `Path` for a hyperbola. The radius of curvature at the apex is specified as `radius`.
    The hyperbola is cut off at the x-coordinate specified by `cutoff`.

    Returns
    -------
    Path
        A `matplotlib.path.Path` object with the parabola segment.
    float
        y-coordinate of the segment's end point.

    Raises
    ------
    ValueError
        If the cutoff is located outside the parabola.

    Arguments
    ---------
    position : float
        Coordinate of the hyperbola apex.
    radius : float
        Radius of curvature at the apex.
    conic : float
        Conic constant (asphericity) of the hyperbola. Must be < -1.
    cutoff
        x-coordinate at which the hyperbola is cut off.
    """
    a, b = _get_hyperbola_sizes(radius, conic)

    # Correct position for hyperbola apex
    position -= a

    if (cutoff < position and radius > 0) or (cutoff > position and radius < 0):
        message = "The cutoff coordinate is located outside the domain of the hyperbola."
        raise ValueError(message)

    t_max = np.arccosh((cutoff - position) / a)
    t = np.linspace(-t_max, t_max, 1000)

    vertices = list(zip(position + a * np.cosh(t), b * np.sinh(t)))
    codes = [Path.MOVETO] + [Path.LINETO] * (len(vertices) - 1)

    return Path(vertices, codes)


def _ellipse(x, rx, ry) -> float:
    """Upper segment of an ellipse. Used to find intersections between lens surfaces."""

    # Make sure the upper half is calculated
    ry = abs(ry)

    return ry * np.sqrt(1 - x**2 / rx**2)


def _parabola(x, rx) -> float:
    """Upper segment of a parabola. Used to find intersections between lens surfaces."""

    return 2 * np.sqrt(rx * x)


def _hyperbola(x, rx, ry) -> float:
    """Upper segment of a hyperbola. Used to find intersections between lens surfaces."""

    # Make sure the upper half is calculated
    ry = abs(ry)

    return ry * np.sqrt(x**2 / rx**2 - 1)


def _lens_surface_function(radius, conic, position) -> Callable[[float], float]:
    """Function for the upper segment of a lens surface. Used to find intersections between lens surfaces."""
    if conic < -1:  # Hyperbola
        rx, ry = _get_hyperbola_sizes(radius, conic)
        x0 = position - rx

        def surface_function(x: float) -> float:
            return _hyperbola(x - x0, rx, ry)

    elif conic == -1:  # Parabola
        rx = radius / 2
        x0 = position

        def surface_function(x: float) -> float:
            return _parabola(x - x0, rx)

    else:  # Ellipse (conic > -1)
        rx, ry = _get_ellipse_sizes(radius, conic)
        x0 = position + rx

        def surface_function(x: float) -> float:
            return _ellipse(x - x0, rx, ry)

    return surface_function


def plot_eye(
    ax: Axes,
    geometry: EyeModel | EyeGeometry,
    lens_edge_thickness: float = 0.0,
    retina_cutoff_position: float | None = None,
    **kwargs,
) -> Axes:
    """Plot an eye.

    Plot an eye with geometric parameters specified by an `EyeGeometry` object.
    The eye is oriented along the horizontal axis, with the pupil center located at `(0, 0)`.
    Additional translations and rotations can be applied using matplotlib patch transforms.

    Parameters
    ----------
    ax : matplotlib.pyplot.Axes
        Matplotlib axes on which the eye will be drawn.
    geometry : EyeGeometry
        Specification of the eye's geometrical parameters.
    lens_edge_thickness : float
        Thickness of the lens at its edge, defaults to 0. If specified, the lens will be cut off at the point where
        its thickness equals this value.
    retina_cutoff_position : float
        Location to which the retina should be drawn. Defaults to the lens's posterior apex.

    Returns
    -------
    matplotlib.pyplot.Axes
        Modified axes with the eye plot

    Raises
    ------
    ValueError
        If `lens_edge_thickness` is less than 0 or `retina_cutoff_position` is located behind the retina.
    """
    # Extract geometry when an EyeModel is supplied
    geometry = geometry if isinstance(geometry, EyeGeometry) else geometry.geometry

    # Input data validation
    if lens_edge_thickness < 0:
        message = f"lens_edge_thickness should be a positive number, got {lens_edge_thickness}."
        raise ValueError(message)
    if (
        retina_cutoff_position is not None
        and retina_cutoff_position > geometry.lens_thickness + geometry.vitreous_thickness
    ):
        message = "retina_cutoff_position is located behind the retina."
        raise ValueError(message)

    # Cornea front
    cornea_front, cornea_cutoff_y = plot_surface(
        -1 * (geometry.cornea_thickness + geometry.anterior_chamber_depth),
        geometry.cornea_front.radius,
        geometry.cornea_front.asphericity,
        cutoff=0,
        return_endpoint=True,
    )

    # Cornea back
    cornea_back = plot_surface(
        -geometry.anterior_chamber_depth,
        geometry.cornea_back.radius,
        geometry.cornea_back.asphericity,
        cutoff=0,
    )

    # Retina
    retina = plot_surface(
        geometry.lens_thickness + geometry.vitreous_thickness,
        geometry.retina.radius,
        geometry.retina.asphericity,
        cutoff=geometry.lens_thickness if retina_cutoff_position is None else retina_cutoff_position,
    )

    # Solve with the lens front surface shifted forward by lens_edge_thickness
    lens_front_function = _lens_surface_function(
        geometry.lens_front.radius,
        geometry.lens_front.asphericity,
        0 + lens_edge_thickness,
    )
    lens_back_function = _lens_surface_function(
        geometry.lens_back.radius,
        geometry.lens_back.asphericity,
        geometry.lens_thickness,
    )

    lens_intersection = fsolve(
        lambda x: lens_front_function(x) - lens_back_function(x),
        x0=geometry.lens_thickness / 2,
    )[0]

    # Lens front
    lens_front = plot_surface(
        0,
        geometry.lens_front.radius,
        geometry.lens_front.asphericity,
        cutoff=lens_intersection - lens_edge_thickness,
    )

    # Lens back
    lens_back = plot_surface(
        geometry.lens_thickness,
        geometry.lens_back.radius,
        geometry.lens_back.asphericity,
        cutoff=lens_intersection,
    )

    # Lens edges
    if lens_edge_thickness > 0:
        codes = [Path.MOVETO, Path.LINETO, Path.MOVETO, Path.LINETO]
        vertices = [
            lens_front.vertices[0],
            lens_front.vertices[0] + (lens_edge_thickness, 0),
            lens_front.vertices[-1],
            lens_front.vertices[-1] + (lens_edge_thickness, 0),
        ]
        lens_edges = Path(vertices, codes)
    else:
        lens_edges = Path(np.zeros((0, 2)))

    # Iris
    codes = [
        Path.MOVETO,
        Path.LINETO,
        Path.MOVETO,
        Path.LINETO,
    ]
    vertices = [
        (0, cornea_cutoff_y),
        (0, geometry.pupil.semi_diameter),
        (0, -geometry.pupil.semi_diameter),
        (0, -cornea_cutoff_y),
    ]
    iris = Path(vertices, codes)

    eye = Path.make_compound_path(cornea_front, cornea_back, iris, lens_front, lens_back, lens_edges, retina)
    ax.add_patch(patches.PathPatch(eye, fill=None, **kwargs))

    return ax



"""Optical properties of materials used in eye models."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable

__all__ = (
    "EyeMaterials",
    "MaterialModel",
    "NavarroMaterials",
    "NavarroMaterials458",
    "NavarroMaterials543",
    "NavarroMaterials589",
    "NavarroMaterials633",
)


@dataclass
class MaterialModel:
    """Optical properties of a material.

    Attributes
    ----------
    refractive_index : float
        The refractive index of the material.
    abbe_number : float
        The Abbe number (or V number) of the material, which is a measure of the material's dispersion
        (variation of refractive index with wavelength).
    partial_dispersion : float
        The partial dispersion of the material, which is the difference in the refractive index of the material
        at two specific wavelengths.
    """

    refractive_index: float
    abbe_number: float
    partial_dispersion: float


@dataclass
class EyeMaterials:
    """Material parameters of the elements of an eye.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.
    """

    cornea: MaterialModel | str
    aqueous: MaterialModel | str
    lens: MaterialModel | str
    vitreous: MaterialModel | str


def _material_model_factory(refractive_index, abbe_number, partial_dispersion) -> Callable[[], MaterialModel]:
    def factory() -> MaterialModel:
        return MaterialModel(refractive_index, abbe_number, partial_dispersion)

    return factory


@dataclass
class NavarroMaterials(EyeMaterials):
    """Material parameters of an eye, according to the Navarro model [1]_.

    The Navarro model defines refractive indices at various wavelengths.
    These data were used to fit the material models in OpticStudio.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3760893927086157,
            abbe_number=53.049630674367904,
            partial_dispersion=0,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.337454140357322,
            abbe_number=48.49905307541748,
            partial_dispersion=0,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.4201120875715691,
            abbe_number=46.3823857404518,
            partial_dispersion=0,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3360870349145728,
            abbe_number=49.685360416317124,
            partial_dispersion=0,
        )
    )


@dataclass
class NavarroMaterials458(EyeMaterials):
    """Material parameters of an eye for a wavelength of 458 nm, according to the Navarro model [1]_.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3828,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3445,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.4292,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3428,
            abbe_number=0,
            partial_dispersion=0,
        )
    )


@dataclass
class NavarroMaterials543(EyeMaterials):
    """Material parameters of an eye for a wavelength of 543 nm, according to the Navarro model [1]_.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3777,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3391,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.4222,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3377,
            abbe_number=0,
            partial_dispersion=0,
        )
    )


@dataclass
class NavarroMaterials589(EyeMaterials):
    """Material parameters of an eye for a wavelength of 589.3 nm, according to the Navarro model [1]_.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.376,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3374,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.42,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.336,
            abbe_number=0,
            partial_dispersion=0,
        )
    )


@dataclass
class NavarroMaterials633(EyeMaterials):
    """Material parameters of an eye for a wavelength of 632.8 nm, according to the Navarro model [1]_.

    Attributes
    ----------
    cornea : MaterialModel
        Refractive model of the cornea.
    aqueous : MaterialModel
        Refractive model of the aqueous humour.
    lens : MaterialModel
        Refractive model of the crystalline lens.
    vitreous : MaterialModel
        Refractive model of the vitreous humour.

    References
    ----------
    .. [1] Escudero-Sanz, I., & Navarro, R. (1999).
       Off-axis aberrations of a wide-angle schematic eye model.
       JOSA A, 16(8), 1881-1891. https://doi.org/10.1364/JOSAA.16.001881
    """

    cornea: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3747,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    aqueous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.336,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    lens: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.4183,
            abbe_number=0,
            partial_dispersion=0,
        )
    )
    vitreous: MaterialModel = field(
        default_factory=_material_model_factory(
            refractive_index=1.3347,
            abbe_number=0,
            partial_dispersion=0,
        )
    )



"""Zernike coefficients analysis for OpticStudio."""

from __future__ import annotations

from typing import TYPE_CHECKING

import zospy as zp

from visisipy.types import SampleSize
from visisipy.wavefront import ZernikeCoefficients

if TYPE_CHECKING:
    from zospy.analyses.wavefront.zernike_standard_coefficients import ZernikeStandardCoefficientsResult

    from visisipy.opticstudio.backend import OpticStudioBackend
    from visisipy.types import FieldCoordinate, FieldType


def _build_zernike_result(zernike_result: ZernikeStandardCoefficientsResult, maximum_term: int) -> ZernikeCoefficients:
    return ZernikeCoefficients({k: v.value for k, v in zernike_result.coefficients.items() if k <= maximum_term})


def zernike_standard_coefficients(
    backend: type[OpticStudioBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int = 45,
) -> tuple[ZernikeCoefficients, ZernikeStandardCoefficientsResult]:
    """Calculate the Zernike standard coefficients at the retina surface.

    Parameters
    ----------
    backend : type[OpticStudioBackend]
        Reference to the OpticStudio backend.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate for the Zernike calculation. When `None`, the first field in OpticStudio is used.
        Defaults to `None`.
    wavelength : float | None, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in OpticStudio is used.
        Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 512.
    maximum_term : int, optional
        The maximum term for the Zernike calculation. Defaults to 45.

    Returns
    -------
    ZernikeCoefficients
        ZOSPy Zernike standard coefficients analysis output.
    """
    if not isinstance(sampling, SampleSize):
        sampling = SampleSize(sampling)

    wavelength_number = 1 if wavelength is None else backend.get_wavelength_number(wavelength)

    if wavelength_number is None:
        backend.set_wavelengths([wavelength])
        wavelength_number = 1

    if field_coordinate is not None:
        backend.set_fields([field_coordinate], field_type=field_type)

    zernike_result = zp.analyses.wavefront.ZernikeStandardCoefficients(
        sampling=str(sampling),
        maximum_term=maximum_term,
        wavelength=wavelength_number,
        field=1,
        reference_opd_to_vertex=False,
        surface="Image",
        sx=0.0,
        sy=0.0,
        sr=1.0,
    ).run(backend.get_oss())

    return _build_zernike_result(zernike_result.data, maximum_term), zernike_result.data



from __future__ import annotations

import pytest

from visisipy.opticstudio import OpticStudioEye

# pyright: reportOptionalMemberAccess=false

pytestmark = [pytest.mark.needs_opticstudio]


class TestOpticStudioEye:
    def test_init(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)

        assert opticstudio_eye.eye_model == eye_model

    def test_build_cornea_front(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.cornea_front.thickness == pytest.approx(eye_model.geometry.cornea_front.thickness)
        assert opticstudio_eye.cornea_front.radius == pytest.approx(eye_model.geometry.cornea_front.radius)
        assert opticstudio_eye.cornea_front.conic == pytest.approx(eye_model.geometry.cornea_front.asphericity)

    def test_build_cornea_back(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.cornea_back.thickness == pytest.approx(eye_model.geometry.cornea_back.thickness)
        assert opticstudio_eye.cornea_back.radius == pytest.approx(eye_model.geometry.cornea_back.radius)
        assert opticstudio_eye.cornea_back.conic == pytest.approx(eye_model.geometry.cornea_back.asphericity)

    def test_build_pupil(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.pupil.semi_diameter == pytest.approx(eye_model.geometry.pupil.semi_diameter)
        assert opticstudio_eye.pupil.is_stop

    def test_build_lens_front(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.lens_front.thickness == pytest.approx(eye_model.geometry.lens_front.thickness)
        assert opticstudio_eye.lens_front.radius == pytest.approx(eye_model.geometry.lens_front.radius)
        assert opticstudio_eye.lens_front.conic == pytest.approx(eye_model.geometry.lens_front.asphericity)

    def test_build_lens_back(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.lens_back.thickness == pytest.approx(eye_model.geometry.lens_back.thickness)
        assert opticstudio_eye.lens_back.radius == pytest.approx(eye_model.geometry.lens_back.radius)
        assert opticstudio_eye.lens_back.conic == pytest.approx(eye_model.geometry.lens_back.asphericity)

    def test_build_retina(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        assert opticstudio_eye.retina.radius == pytest.approx(eye_model.geometry.retina.radius)
        assert opticstudio_eye.retina.conic == pytest.approx(eye_model.geometry.retina.asphericity)
        assert opticstudio_eye.retina.surface.IsImage

    def test_build_pupil_not_stop_raises_valueerror(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.lens_front._is_stop = True

        with pytest.raises(ValueError, match="The pupil is not located at the stop position"):
            opticstudio_eye.build(oss)

    def test_build_retina_not_image_raises_valueerror(self, oss, eye_model):
        oss.LDE.InsertNewSurfaceAt(oss.LDE.NumberOfSurfaces)
        opticstudio_eye = OpticStudioEye(eye_model)

        with pytest.raises(ValueError, match="The retina is not located at the image position"):
            opticstudio_eye.build(oss)

    def test_build_start_from_index(self, oss, eye_model):
        oss.LDE.InsertNewSurfaceAt(1).Comment = "dummy"

        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss, start_from_index=1)

        assert opticstudio_eye.cornea_front.surface.SurfaceNumber == 2
        assert opticstudio_eye.cornea_back.surface.SurfaceNumber == 3
        assert opticstudio_eye.pupil.surface.SurfaceNumber == 4
        assert opticstudio_eye.lens_front.surface.SurfaceNumber == 5
        assert opticstudio_eye.lens_back.surface.SurfaceNumber == 6
        assert opticstudio_eye.retina.surface.SurfaceNumber == 7

    def test_build_start_from_index_after_stop_surface_raises_valueerror(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)

        with pytest.raises(ValueError, match="'start_from_index' must be smaller than the index of the stop surface"):
            opticstudio_eye.build(oss, start_from_index=1)

    def test_build_replace_existing(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        eye_model.geometry.cornea_front.thickness = 0.1
        eye_model.geometry.lens_front.radius = 12

        new_opticstudio_eye = OpticStudioEye(eye_model)
        new_opticstudio_eye.build(oss, replace_existing=True)

        assert oss.LDE.GetSurfaceAt(1).Thickness == 0.1
        assert oss.LDE.GetSurfaceAt(4).Radius == 12

    def test_build_object_distance(self, oss, eye_model):
        assert oss.LDE.GetSurfaceAt(0).Thickness == float("inf")

        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss, object_distance=1.0)

        assert oss.LDE.GetSurfaceAt(0).Thickness == 1.0

    def test_set_cornea_front_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.cornea_front.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(1).Comment == "new comment"

    def test_set_cornea_back_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.cornea_back.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(2).Comment == "new comment"

    def test_set_pupil_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.pupil.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(3).Comment == "new comment"

    def test_set_lens_front_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.lens_front.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(4).Comment == "new comment"

    def test_set_lens_back_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.lens_back.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(5).Comment == "new comment"

    def test_set_retina_property(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.retina.comment = "new comment"

        assert oss.LDE.GetSurfaceAt(6).Comment == "new comment"

    @pytest.mark.parametrize("build", [True, False])
    def test_surfaces(self, oss, eye_model, build):
        opticstudio_eye = OpticStudioEye(eye_model)

        if build:
            opticstudio_eye.build(oss)

        assert opticstudio_eye.surfaces == {
            "cornea_front": opticstudio_eye.cornea_front,
            "cornea_back": opticstudio_eye.cornea_back,
            "pupil": opticstudio_eye.pupil,
            "lens_front": opticstudio_eye.lens_front,
            "lens_back": opticstudio_eye.lens_back,
            "retina": opticstudio_eye.retina,
        }

    def test_update_surfaces(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.update_surfaces("comment", "new comment")

        assert opticstudio_eye.cornea_front.comment == "new comment"
        assert opticstudio_eye.cornea_back.comment == "new comment"
        assert opticstudio_eye.pupil.comment == "new comment"
        assert opticstudio_eye.lens_front.comment == "new comment"
        assert opticstudio_eye.lens_back.comment == "new comment"
        assert opticstudio_eye.retina.comment == "new comment"

    def test_update_surfaces_subset(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        opticstudio_eye.update_surfaces("comment", "new comment", surface_names=["cornea_front", "lens_back"])

        assert opticstudio_eye.cornea_front.comment == "new comment"
        assert opticstudio_eye.cornea_back.comment == "cornea back / aqueous"
        assert opticstudio_eye.pupil.comment == "pupil"
        assert opticstudio_eye.lens_front.comment == "lens front"
        assert opticstudio_eye.lens_back.comment == "new comment"
        assert opticstudio_eye.retina.comment == "retina"

    def test_relink_surfaces(self, oss, eye_model):
        opticstudio_eye = OpticStudioEye(eye_model)
        opticstudio_eye.build(oss)

        oss.LDE.InsertNewSurfaceAt(1).Comment = "dummy"

        assert opticstudio_eye.relink_surfaces(oss)
        assert opticstudio_eye.cornea_front.surface.SurfaceNumber == 2
        assert opticstudio_eye.cornea_front.comment == "cornea front"
        assert opticstudio_eye.cornea_back.surface.SurfaceNumber == 3
        assert opticstudio_eye.cornea_back.comment == "cornea back / aqueous"
        assert opticstudio_eye.pupil.surface.SurfaceNumber == 4
        assert opticstudio_eye.pupil.comment == "pupil"
        assert opticstudio_eye.lens_front.surface.SurfaceNumber == 5
        assert opticstudio_eye.lens_front.comment == "lens front"
        assert opticstudio_eye.lens_back.surface.SurfaceNumber == 6
        assert opticstudio_eye.lens_back.comment == "lens back / vitreous"
        assert opticstudio_eye.retina.surface.SurfaceNumber == 7
        assert opticstudio_eye.retina.comment == "retina"


def test_model_with_no_surface(oss, three_surface_eye_model):
    opticstudio_eye = OpticStudioEye(three_surface_eye_model)
    opticstudio_eye.build(oss)

    assert oss.LDE.NumberOfSurfaces == 6
    assert opticstudio_eye.cornea_front.surface.SurfaceNumber == 1
    assert opticstudio_eye.pupil.surface.SurfaceNumber == 2
    assert opticstudio_eye.lens_front.surface.SurfaceNumber == 3
    assert opticstudio_eye.lens_back.surface.SurfaceNumber == 4
    assert opticstudio_eye.retina.surface.SurfaceNumber == 5
    assert opticstudio_eye.cornea_back.surface is None



"""Calculate the Zernike standard coefficients for an eye model."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

import numpy as np

from visisipy.analysis.base import _AUTOMATIC_BACKEND, analysis
from visisipy.types import FieldType, SampleSize
from visisipy.wavefront import min_max_noll_index

if TYPE_CHECKING:
    from visisipy.backend import BaseBackend
    from visisipy.models import EyeModel
    from visisipy.wavefront import ZernikeCoefficients

__all__ = (
    "rms_hoa",
    "zernike_standard_coefficients",
)


@overload
def zernike_standard_coefficients(
    model: EyeModel | None = None,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int = 45,
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> ZernikeCoefficients: ...


@overload
def zernike_standard_coefficients(
    model: EyeModel | None = None,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int = 45,
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[ZernikeCoefficients, Any]: ...


@analysis
def zernike_standard_coefficients(
    model: EyeModel | None = None,  # noqa: ARG001
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: Literal["angle", "object_height"] = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int = 45,
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[ZernikeCoefficients, Any]:
    """Calculates the Zernike standard coefficients at the retina surface.

    Zernike standard coefficients are returned in the Noll notation.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate for the Zernike calculation. When `None`, the first field in the backend is used.
        Defaults to `None`.
    wavelength : float | None, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in the backend is used.
        Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 64.
    maximum_term : int, optional
        The maximum term for the Zernike calculation. Defaults to 45.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the current backend is used.

    Returns
    -------
    ZernikeCoefficients
        Zernike standard coefficients in Noll notation.
    """
    return backend.analysis.zernike_standard_coefficients(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=SampleSize(sampling),
        maximum_term=maximum_term,
    )


@overload
def rms_hoa(
    model: EyeModel | None = None,
    min_order: int = 3,
    max_order: int = 6,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int | None = None,
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> float: ...


@overload
def rms_hoa(
    model: EyeModel | None = None,
    min_order: int = 3,
    max_order: int = 6,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int | None = None,
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[float, Any]: ...


@analysis
def rms_hoa(
    model: EyeModel | None = None,  # noqa: ARG001
    min_order: int = 3,
    max_order: int = 8,
    field_coordinate: tuple[float, float] | None = None,
    wavelength: float | None = None,
    field_type: FieldType = "angle",
    sampling: SampleSize | str | int = 64,
    maximum_term: int | None = None,
    *,
    return_raw_result: bool = True,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[float, Any]:
    """Calculates the root-mean-square (RMS) of higher-order aberrations (HOA) in the eye model.

    By default, the Zernike orders from 3 to 8 are used. The RMS is calculated following the definition in the ANSI
    Z80.28-2010 standard [1]_:

    .. math:: RMS_{WFE} = \\sqrt{ \\sum_{n>1, all\\ m} (C_{n}^{m})^2 }

    .. [1]: VC ANSI Z80.28-2010—Ophthalmics—Methods of Reporting Optical Aberrations of Eyes. (n.d.).
            https://webstore.ansi.org/standards/ansi/vcansiz80282010

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    min_order : int
        The minimum Zernike polynomial order to be included in the calculation. Defaults to 3.
    max_order : int
        The maximum Zernike polynomial order to be included in the calculation. Defaults to 8.
    field_coordinate : tuple[float, float] | None, optional
        The field coordinate for the Zernike calculation. When `None`, the first field in the backend is used.
        Defaults to `None`.
    wavelength : float | None, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in the backend is used.
        Defaults to `None`.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 64.
    maximum_term : int | None, optional
        The maximum term for the Zernike calculation. If `None`, the maximum term is set to the largest term of
        `max_order`. Defaults to `None`.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the current backend is used.

    Returns
    -------
    float
        The root-mean-square (RMS) of higher-order aberrations (HOA) in the eye model.

    Raises
    ------
    ValueError
        If `min_order` or `max_order` is less than 0.
        If `max_order` is less than or equal to `min_order`.
        If `maximum_term` is less than the largest term of `max_order`.
    """
    if min_order < 0 or max_order < 0:
        raise ValueError("min_order and max_order must be greater than or equal to 0")

    if max_order <= min_order:
        raise ValueError("max_order must be greater than min_order")

    min_index, max_index = min_max_noll_index(min_order, max_order)

    if maximum_term is None:
        maximum_term = max_index
    elif maximum_term < max_index:
        raise ValueError(f"maximum_term must be greater than or equal to the largest term of max_order: {max_index}")

    zernikes, raw_result = backend.analysis.zernike_standard_coefficients(
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=SampleSize(sampling),
        maximum_term=maximum_term,
    )

    rms_aberrations = float(np.sqrt(sum(zernikes[i] ** 2 for i in range(min_index, max_index + 1))))

    return rms_aberrations, raw_result



"""Ray trace analysis for Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from collections.abc import Iterable

    from optiland.optic import Optic

    from visisipy.backend import FieldCoordinate, FieldType
    from visisipy.optiland import OptilandBackend

__all__ = ("raytrace",)


def _trace_single_ray(
    optic: Optic,
    field: FieldCoordinate,
    pupil: tuple[float, float],
    wavelength: float,
) -> pd.DataFrame:
    optic.trace_generic(*field, *pupil, wavelength=wavelength)
    x = optic.surface_group.x
    y = optic.surface_group.y
    z = optic.surface_group.z
    surface_numbers = range(optic.surface_group.num_surfaces)

    return pd.DataFrame(
        {
            "wavelength": [wavelength] * len(x),
            "surface": surface_numbers,
            "comment": [None] * len(x),
            "x": x[:, 0],
            "y": y[:, 0],
            "z": z[:, 0],
        }
    )


def raytrace(
    backend: type[OptilandBackend],
    coordinates: Iterable[FieldCoordinate] | None = None,
    wavelengths: Iterable[float] | None = None,
    field_type: FieldType = "angle",
    pupil: tuple[float, float] = (0, 0),
) -> tuple[pd.DataFrame, None]:
    """Perform a ray trace analysis using the given parameters.

    The ray trace is performed for each wavelength and field in the system, using the generic ray trace in Optiland.

    The analysis returns a Dataframe with the following columns:

    - field: The field coordinates for the ray trace.
    - wavelength: The wavelength used in the ray trace.
    - surface: The surface number in the system.
    - comment: The comment for the surface.
    - x: The X-coordinate of the ray trace.
    - y: The Y-coordinate of the ray trace.
    - z: The Z-coordinate of the ray trace.

    Parameters
    ----------
    backend : type[OptilandBackend]
        The Optiland backend to use for the ray trace.
    coordinates : Iterable[tuple[float, float]], optional
        An iterable of tuples representing the coordinates for the ray trace.
        If `field_type` is "angle", the coordinates should be the angles along the (X, Y) axes in degrees.
        If `field_type` is "object_height", the coordinates should be the object heights along the
        (X, Y) axes in mm. Defaults to `None`, which uses the fields defined in the backend.
    wavelengths : Iterable[float], optional
        An iterable of wavelengths to be used in the ray trace. Defaults to `None`, which uses the wavelengths
        defined in the backend.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used in the ray trace. Can be either "angle" or "object_height". Defaults to
        "angle".
    pupil : tuple[float, float], optional
        A tuple representing the pupil coordinates for the ray trace. Defaults to (0, 0).

    Returns
    -------
    DataFrame
        A pandas DataFrame containing the results of the ray trace analysis.
    """
    if abs(pupil[0]) > 1 or abs(pupil[1]) > 1:
        raise ValueError("Pupil coordinates must be between -1 and 1.")

    if coordinates is not None:
        backend.set_fields(coordinates, field_type)

    if wavelengths is not None:
        backend.set_wavelengths(wavelengths)

    raytrace_results = []

    normalized_fields = backend.get_optic().fields.get_field_coords()

    for _, wavelength in backend.iter_wavelengths():
        for (_, field), normalized_field in zip(backend.iter_fields(), normalized_fields):
            result = _trace_single_ray(backend.get_optic(), normalized_field, pupil, wavelength)
            result.insert(0, "field", [field] * len(result))
            raytrace_results.append(result)

    return pd.concat(raytrace_results).reset_index(), None



from __future__ import annotations

from types import SimpleNamespace
from typing import TYPE_CHECKING

import optiland.materials
import pytest

from visisipy.models.geometry import NoSurface, StandardSurface, Stop, Surface
from visisipy.models.materials import MaterialModel
from visisipy.optiland.surfaces import (
    OptilandNoSurface,
    OptilandSurface,
    _built_only_property,
    make_surface,
)

if TYPE_CHECKING:
    import optiland.surfaces  # noqa: TC004
    from optiland.optic import Optic


@pytest.fixture
def simple_system(optic: Optic) -> Optic:
    """Simple system with an object and image surface."""
    optic.add_surface(
        index=0,
        thickness=float("inf"),
        comment="Object",
    )

    optic.add_surface(
        index=1,
        radius=float("inf"),
        thickness=0,
        conic=0.0,
        comment="Image",
    )

    return optic


@pytest.fixture
def surface(simple_system: Optic) -> optiland.surfaces.Surface:
    """Fixture to create a surface for testing."""
    simple_system.add_surface(
        index=1,
        radius=1.0,
        thickness=2.0,
        conic=0.0,
        material="BK7",
        is_stop=True,
        comment="Test surface",
    )

    return simple_system.surface_group.surfaces[1]


class TestBuiltOnlyProperty:
    MockSurface = SimpleNamespace(
        comment="Test surface",
    )

    class MockOptilandSurface:
        surface: SimpleNamespace | None = None

        def __init__(self, surface) -> None:
            self.surface = surface
            self._is_built = False

        @_built_only_property
        def comment(self) -> str:
            return self.surface.comment  # type: ignore

        @comment.setter
        def comment(self, value: str) -> None:
            self.surface.comment = value  # type: ignore

        def build(self) -> None:
            self._is_built = True

    def test_get_built(self):
        """Test getting a built property."""
        mock_surface = self.MockOptilandSurface(self.MockSurface)
        mock_surface.build()

        assert mock_surface.comment == "Test surface"

    def test_get_not_built(self):
        """Test getting a property before it is built."""
        mock_surface = self.MockOptilandSurface(self.MockSurface)

        assert mock_surface.comment is None

    def test_set_built(self):
        """Test setting a built property."""
        mock_surface = self.MockOptilandSurface(self.MockSurface)
        mock_surface.build()

        mock_surface.comment = "New comment"

        assert mock_surface.surface.comment == "New comment"  # pyright: ignore [reportOptionalMemberAccess]

    def test_set_not_built(self):
        """Test setting a property before it is built."""
        mock_surface = self.MockOptilandSurface(self.MockSurface)

        with pytest.raises(AttributeError, match="Cannot set attribute of non-built surface"):
            mock_surface.comment = "New comment"


def build_surface(optic: Optic, surface: OptilandSurface) -> int:
    """Build a surface with `surface_settings` to the optic system, between an object and image surface."""
    optic.add_surface(
        index=0,
        thickness=float("inf"),
        comment="Object",
    )

    index = surface.build(optic, position=1)

    optic.add_surface(
        index=2,
        radius=float("inf"),
        thickness=0,
        conic=0.0,
        comment="Image",
    )

    return index


class TestOptilandSurface:
    @pytest.mark.parametrize("is_stop", [True, False, None])
    def test_build(self, optic: Optic, is_stop: bool):
        surface = OptilandSurface(
            comment="Test surface",
            radius=1.0,
            thickness=2.0,
            semi_diameter=1.0,
            conic=0.0,
            material="BK7",
            is_stop=is_stop,
        )

        assert surface._is_built is False
        assert surface.surface is None
        assert surface._optic is None
        assert surface._index is None

        surface_index = build_surface(optic, surface)

        assert surface_index == 1
        assert surface.surface is not None
        assert surface._is_built is True
        assert surface._optic == optic
        assert surface._index == 1

        assert surface.comment == "Test surface"
        assert surface.radius == 1.0
        assert surface.thickness == 2.0
        assert surface.semi_diameter == 1.0
        assert surface.conic == 0.0
        assert surface.material == "BK7"
        assert surface.is_stop == bool(is_stop)

    def test_build_ideal_material(self, optic: Optic):
        material = MaterialModel(refractive_index=1.5, abbe_number=0.0, partial_dispersion=0.0)
        surface = OptilandSurface(
            comment="Test surface",
            radius=1.0,
            thickness=2.0,
            semi_diameter=1.0,
            conic=0.0,
            material=material,
            is_stop=True,
        )

        surface_index = build_surface(optic, surface)

        assert surface_index == 1
        assert surface.material == material
        assert isinstance(surface.surface.material_post, optiland.materials.IdealMaterial)
        assert surface.surface.material_post.index == material.refractive_index
        assert surface.surface.material_post.k(0.543) == 0

    def test_build_abbe_material(self, optic: Optic):
        material = MaterialModel(refractive_index=1.5, abbe_number=50.0, partial_dispersion=0.0)
        surface = OptilandSurface(
            comment="Test surface",
            radius=1.0,
            thickness=2.0,
            semi_diameter=1.0,
            conic=0.0,
            material=material,
            is_stop=True,
        )

        surface_index = build_surface(optic, surface)

        assert surface_index == 1
        assert surface.material == material
        assert isinstance(surface.surface.material_post, optiland.materials.AbbeMaterial)
        assert surface.surface.material_post.index == material.refractive_index
        assert surface.surface.material_post.abbe == material.abbe_number

    def test_build_str_material(self, optic: Optic):
        material = "BK7"
        surface = OptilandSurface(
            comment="Test surface",
            radius=1.0,
            thickness=2.0,
            semi_diameter=1.0,
            conic=0.0,
            material=material,
            is_stop=True,
        )

        surface_index = build_surface(optic, surface)

        assert surface_index == 1
        assert surface.material == material
        assert isinstance(surface.surface.material_post, optiland.materials.Material)
        assert surface.surface.material_post.name == material

    def test_set_material_incorrect_type_raises_typeerror(self, optic: Optic):
        surface = OptilandSurface(comment="Test surface", material=123)

        with pytest.raises(TypeError, match="'material' must be MaterialModel or str"):
            build_surface(optic, surface)


class TestNoSurface:
    def test_build(self, optic):
        surface = OptilandNoSurface()

        n_surfaces = optic.surface_group.num_surfaces

        return_index = surface.build(optic, position=1)

        assert return_index == 0
        assert surface.surface is None
        assert n_surfaces == optic.surface_group.num_surfaces


class TestMakeSurface:
    def test_make_surface(self):
        surface = Surface(thickness=1)

        optiland_surface = make_surface(surface, material="BK7")

        assert optiland_surface._thickness == 1
        assert optiland_surface._material == "BK7"

    @pytest.mark.parametrize(
        "radius,thickness,semi_diameter,asphericity,material",
        [
            (1, 2, 3, 4, "BK7"),
            (
                5,
                6,
                7,
                8,
                MaterialModel(refractive_index=1.5, abbe_number=50, partial_dispersion=0.67),
            ),
        ],
    )
    def test_make_standard_surface(self, radius, thickness, semi_diameter, asphericity, material):
        surface = StandardSurface(
            radius=radius,
            thickness=thickness,
            semi_diameter=semi_diameter,
            asphericity=asphericity,
        )

        optiland_surface = make_surface(surface, material)

        assert optiland_surface._radius == radius
        assert optiland_surface._thickness == thickness
        assert optiland_surface._semi_diameter == semi_diameter
        assert optiland_surface._conic == asphericity
        assert optiland_surface._material == material

    @pytest.mark.parametrize(
        "thickness,semi_diameter,material",
        [
            (1, 2, "BK7"),
            (
                3,
                4,
                MaterialModel(refractive_index=1.5, abbe_number=50, partial_dispersion=0.67),
            ),
        ],
    )
    def test_make_stop_surface(self, thickness, semi_diameter, material):
        surface = Stop(thickness=thickness, semi_diameter=semi_diameter)

        optiland_surface = make_surface(surface, material)

        assert optiland_surface._thickness == thickness
        assert optiland_surface._semi_diameter == semi_diameter
        assert optiland_surface._material == material
        assert optiland_surface._is_stop is True

    def test_make_no_surface(self):
        surface = NoSurface()
        opticstudio_surface = make_surface(surface)

        assert isinstance(opticstudio_surface, OptilandNoSurface)



"""Cardinal points analysis for Optiland."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

from visisipy.analysis.cardinal_points import CardinalPoints, CardinalPointsResult

if TYPE_CHECKING:
    from optiland.paraxial import Paraxial

    from visisipy.optiland import OptilandBackend


__all__ = ("cardinal_points",)


def _build_cardinal_points_result(paraxial: Paraxial) -> CardinalPointsResult:
    return CardinalPointsResult(
        focal_lengths=CardinalPoints(
            object=paraxial.f1(),
            image=paraxial.f2(),
        ),
        focal_points=CardinalPoints(
            object=paraxial.F1(),
            image=paraxial.F2(),
        ),
        principal_points=CardinalPoints(
            object=paraxial.P1(),
            image=paraxial.P2(),
        ),
        anti_principal_points=CardinalPoints(
            object=paraxial.P1anti(),
            image=paraxial.P2anti(),
        ),
        nodal_points=CardinalPoints(
            object=paraxial.N1(),
            image=paraxial.N2(),
        ),
        anti_nodal_points=CardinalPoints(
            object=paraxial.N1anti(),
            image=paraxial.N2anti(),
        ),
    )


def cardinal_points(
    backend: OptilandBackend, surface_1: int | None = None, surface_2: int | None = None
) -> tuple[CardinalPointsResult, Paraxial]:
    """Get the cardinal points of the system between `surface_1` and `surface_2`.

    Note that Optiland only supports calculating cardinal points for the entire system, not for a subset of surfaces.
    A ValueError will be raised if `surface_1` or `surface_2` are different from the first and last surfaces in the system.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the Optiland backend.
    surface_1 : int | None, optional
        The first surface to be used in the analysis. If `None`, the first surface in the system will be used.
        Defaults to `None`.
    surface_2 : int | None, optional
        The second surface to be used in the analysis. If `None`, the last surface in the system will be used.
        Defaults to `None`.

    Returns
    -------
    CardinalPointsResult
        The cardinal points of the system.

    Raises
    ------
    ValueError
        If `surface_1` or `surface_2` are not between 1 and the number of surfaces in the system, or if `surface_1`
        is greater than or equal to `surface_2`.
    """
    if (surface_1 is not None and surface_1 not in {0, 1}) or (
        surface_2 is not None and surface_2 != backend.get_optic().surface_group.num_surfaces - 1
    ):
        raise ValueError("Optiland only supports calculating cardinal points for the entire system.")

    return _build_cardinal_points_result(backend.get_optic().paraxial), deepcopy(backend.get_optic().paraxial)



from __future__ import annotations

import numpy as np
import pytest
import zospy as zp

from visisipy.models.materials import NavarroMaterials

pytestmark = [pytest.mark.needs_opticstudio]


@pytest.fixture
def opticstudio_model(oss):
    """Simple optical system with 3 parallel surfaces: a stop, refracting interface, and image surface.
    All surfaces except the refracting interface and image surface have a refractive index of 1.0.
    The refractive index of the image surface is equal to that of the refracting interface, to prevent reflection.

    When a single ray trace is performed, the angle of incidence at the image surface is equal to the angle of
    refraction at the refracting interface. This allows to calculate the refractive index of the refracting interface
    from the angle of incidence at the refracting interface and the angle of incidence at the image surface.
    """
    oss.SystemData.Aperture.ApertureType = zp.constants.SystemData.ZemaxApertureType.FloatByStopSize

    oss.SystemData.Fields.SetFieldType(zp.constants.SystemData.FieldType.Angle)
    oss.SystemData.Fields.GetField(1).Y = 20

    obj = oss.LDE.GetSurfaceAt(0)
    obj.Comment = "object"
    obj.Thickness = float("inf")
    zp.solvers.material_model(
        obj.MaterialCell,
        refractive_index=1.0,
        abbe_number=0.0,
        partial_dispersion=0.0,
    )

    stop = oss.LDE.GetSurfaceAt(1)
    stop.Comment = "stop"
    stop.Thickness = 5.0
    stop.SemiDiameter = 1.0
    zp.solvers.material_model(stop.MaterialCell, refractive_index=1.0, abbe_number=0.0, partial_dispersion=0.0)

    interface = oss.LDE.InsertNewSurfaceAt(2)
    interface.Comment = "interface"
    interface.Thickness = 5.0
    zp.solvers.material_model(
        interface.MaterialCell,
        refractive_index=1.5,
        abbe_number=0.0,
        partial_dispersion=0.0,
    )

    image = oss.LDE.GetSurfaceAt(3)
    image.Comment = "image"
    zp.solvers.material_model(
        image.MaterialCell,
        refractive_index=1.5,
        abbe_number=0.0,
        partial_dispersion=0.0,
    )

    return oss


def _set_material_model(material_cell, material_model):
    zp.solvers.material_model(
        material_cell,
        refractive_index=material_model.refractive_index,
        abbe_number=material_model.abbe_number,
        partial_dispersion=material_model.partial_dispersion,
    )


@pytest.mark.parametrize(
    "material_model,wavelength,expected_index",
    [
        # Cornea
        (NavarroMaterials().cornea, 0.458, 1.3828),
        (NavarroMaterials().cornea, 0.543, 1.3777),
        (NavarroMaterials().cornea, 0.5893, 1.376),
        (NavarroMaterials().cornea, 0.6328, 1.3747),
        # Aqueous
        (NavarroMaterials().aqueous, 0.458, 1.3445),
        (NavarroMaterials().aqueous, 0.543, 1.3391),
        (NavarroMaterials().aqueous, 0.5893, 1.3374),
        (NavarroMaterials().aqueous, 0.6328, 1.336),
        # Lens
        (NavarroMaterials().lens, 0.458, 1.4292),
        (NavarroMaterials().lens, 0.543, 1.4222),
        (NavarroMaterials().lens, 0.5893, 1.42),
        (NavarroMaterials().lens, 0.6328, 1.4183),
        # Vitreous
        (NavarroMaterials().vitreous, 0.458, 1.3428),
        (NavarroMaterials().vitreous, 0.543, 1.3377),
        (NavarroMaterials().vitreous, 0.5893, 1.336),
        (NavarroMaterials().vitreous, 0.6328, 1.3347),
    ],
)
def test_material_model_refractive_index(opticstudio_model, material_model, wavelength, expected_index):
    # Set wavelength
    opticstudio_model.SystemData.Wavelengths.GetWavelength(1).Wavelength = wavelength

    # Set material model of test surface and image surface to the material model
    _set_material_model(opticstudio_model.LDE.GetSurfaceAt(2).MaterialCell, material_model)
    _set_material_model(opticstudio_model.LDE.GetSurfaceAt(3).MaterialCell, material_model)

    # Get angle of incidence at image surface. As the refracting surface and the image surface are parallel, the angle
    # of incidence at the image surface is equal to the angle of refraction at the refracting surface.
    raytrace_result = zp.analyses.raysandspots.SingleRayTrace(
        hx=0, hy=1, px=0, py=1, wavelength=1, global_coordinates=True
    ).run(opticstudio_model)
    angle = raytrace_result.data.real_ray_trace_data.loc[3]["Angle in"]
    refractive_index = np.sin(np.deg2rad(20)) / np.sin(np.deg2rad(angle))

    assert refractive_index == pytest.approx(expected_index, rel=1e-3)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.models import EyeModel


class TestCardinalPointsAnalysis:
    @pytest.mark.parametrize(
        "surface_1,surface_2,expectation",
        [
            (None, None, does_not_raise()),
            (1, 6, does_not_raise()),
            (
                2,
                4,
                pytest.raises(
                    ValueError, match="Optiland only supports calculating cardinal points for the entire system"
                ),
            ),
            (
                -1,
                7,
                pytest.raises(
                    ValueError,
                    match="Optiland only supports calculating cardinal points for the entire system",
                ),
            ),
        ],
    )
    def test_cardinal_points(self, surface_1, surface_2, expectation, optiland_backend, optiland_analysis):
        optiland_backend.build_model(EyeModel())

        with expectation:
            assert optiland_analysis.cardinal_points(surface_1, surface_2)



from __future__ import annotations

import math
from contextlib import nullcontext as does_not_raise
from typing import TYPE_CHECKING

import numpy as np
import pytest
import zospy as zp

from visisipy import EyeModel

if TYPE_CHECKING:
    from visisipy.opticstudio.backend import OpticStudioSettings

pytestmark = [pytest.mark.needs_opticstudio]


class TestOpticStudioBackend:
    def test_initialize_opticstudio(self, opticstudio_backend):
        assert opticstudio_backend.zos is not None
        assert opticstudio_backend.oss is not None

    def test_new_model(self, opticstudio_backend):
        # Change a setting and add a new surface
        opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(1).Wavelength = 0.640
        opticstudio_backend.oss.LDE.InsertNewSurfaceAt(2)

        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 4

        opticstudio_backend.new_model(save_old_model=False)

        assert (
            opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(1).Wavelength
            == opticstudio_backend.settings["wavelengths"][0]
        )
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 3

    def test_build_model(self, opticstudio_backend):
        model = EyeModel()

        opticstudio_backend.build_model(model)

        assert opticstudio_backend.model is not None
        assert opticstudio_backend.model.eye_model == model
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 7

    def test_clear_model(self, opticstudio_backend):
        model = EyeModel()

        opticstudio_backend.build_model(model)
        assert opticstudio_backend.model is not None
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 7

        opticstudio_backend.clear_model()

        assert opticstudio_backend.model is None
        assert opticstudio_backend.oss.LDE.NumberOfSurfaces == 3

    def test_save_model(self, opticstudio_backend, tmp_path):
        model = EyeModel()

        opticstudio_backend.build_model(model)

        # Save the model to a temporary path
        model_path = tmp_path / "model.zmx"
        opticstudio_backend.save_model(model_path)

        assert model_path.exists()

        # Remove the model file
        model_path.unlink()

        assert not model_path.exists()

        # Save without specifying a path
        opticstudio_backend.save_model()

        assert model_path.exists()

    def test_disconnect(self, opticstudio_backend):
        opticstudio_backend.disconnect()

        assert opticstudio_backend.zos is None
        assert opticstudio_backend.oss is None

    @pytest.mark.parametrize(
        "coordinates,field_type,field_constant,expectation",
        [
            ([(10, 10)], "angle", "Angle", does_not_raise()),
            (
                [(0, 0), (0, 10), (-10, 0), (10, -10)],
                "angle",
                "Angle",
                does_not_raise(),
            ),
            (
                [(math.pi, 4.15), (-12, 1)],
                "object_height",
                "ObjectHeight",
                does_not_raise(),
            ),
            (
                [(np.int64(10), np.int64(11))],  # Test float conversion
                "angle",
                "Angle",
                does_not_raise(),
            ),
            (
                [(0, 0)],
                "invalid",
                None,
                pytest.raises(
                    ValueError,
                    match="field_type must be either 'angle' or 'object_height'",
                ),
            ),
        ],
    )
    def test_set_fields(self, opticstudio_backend, coordinates, field_type, field_constant, expectation):
        with expectation:
            opticstudio_backend.set_fields(coordinates, field_type)

            assert opticstudio_backend.oss.SystemData.Fields.NumberOfFields == len(coordinates)
            for i in range(len(coordinates)):
                field = opticstudio_backend.oss.SystemData.Fields.GetField(i + 1)
                assert coordinates[i] == (field.X, field.Y)

            assert opticstudio_backend.oss.SystemData.Fields.GetFieldType() == zp.constants.process_constant(
                zp.constants.SystemData.FieldType, field_constant
            )

    def test_get_fields(self, opticstudio_backend):
        coordinates = [(0, 0), (0, 10), (-10, 0), (10, -10)]

        for i, f in enumerate(coordinates, start=1):
            if i == 1:
                field = opticstudio_backend.get_oss().SystemData.Fields.GetField(i)
                field.X, field.Y = f
            else:
                opticstudio_backend.get_oss().SystemData.Fields.AddField(*f, 1)

        assert opticstudio_backend.get_fields() == coordinates

    def test_set_wavelengths(self, opticstudio_backend):
        opticstudio_backend.set_wavelengths([0.543, 0.650])

        assert opticstudio_backend.oss.SystemData.Wavelengths.NumberOfWavelengths == 2
        assert opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(1).Wavelength == 0.543
        assert opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(2).Wavelength == 0.650

    def test_get_wavelengths(self, opticstudio_backend):
        wavelengths = [0.543, 0.650]

        for i, w in enumerate(wavelengths, start=1):
            if i == 1:
                opticstudio_backend.get_oss().SystemData.Wavelengths.GetWavelength(i).Wavelength = w
            else:
                opticstudio_backend.get_oss().SystemData.Wavelengths.AddWavelength(w, i)

        assert opticstudio_backend.get_wavelengths() == wavelengths

    def test_get_wavelength_number(self, opticstudio_backend):
        opticstudio_backend.set_wavelengths([0.543, 0.650])

        assert opticstudio_backend.get_wavelength_number(0.543) == 1
        assert opticstudio_backend.get_wavelength_number(0.650) == 2
        assert opticstudio_backend.get_wavelength_number(1.234) is None


class TestOpticStudioBackendSettings:
    def test_update_settings(self, opticstudio_backend):
        settings: OpticStudioSettings = {
            "field_type": "object_height",
            "fields": [(0, 0), (0, 10), (-10, 0), (10, -10)],
            "ray_aiming": "paraxial",
            "wavelengths": [0.543, 0.650],
            "aperture_type": "entrance_pupil_diameter",
            "aperture_value": 3.0,
        }

        opticstudio_backend.update_settings(**settings)

        assert all(opticstudio_backend.settings[k] == v for k, v in settings.items())

    @pytest.mark.parametrize(
        "field_type,fields,expected_field_type",
        [
            ("angle", [(10, 10)], "Angle"),
            ("object_height", [(0, 0), (0, 10), (-10, 0), (10, -10)], "ObjectHeight"),
        ],
    )
    def test_field(self, field_type, fields, expected_field_type, opticstudio_backend):
        opticstudio_backend.update_settings(field_type=field_type, fields=fields)

        assert opticstudio_backend.oss.SystemData.Fields.NumberOfFields == len(fields)
        for i in range(len(fields)):
            field = opticstudio_backend.oss.SystemData.Fields.GetField(i + 1)
            assert fields[i] == (field.X, field.Y)

        assert opticstudio_backend.oss.SystemData.Fields.GetFieldType() == zp.constants.process_constant(
            zp.constants.SystemData.FieldType, expected_field_type
        )

    @pytest.mark.parametrize(
        "wavelengths",
        [[0.543, 0.650], [0.543, 0.650, 0.450]],
    )
    def test_wavelength(self, wavelengths, opticstudio_backend):
        opticstudio_backend.update_settings(wavelengths=wavelengths)

        assert opticstudio_backend.oss.SystemData.Wavelengths.NumberOfWavelengths == len(wavelengths)
        for i, wavelength in enumerate(wavelengths):
            assert opticstudio_backend.oss.SystemData.Wavelengths.GetWavelength(i + 1).Wavelength == wavelength

    @pytest.mark.parametrize(
        "aperture_type, aperture_value, expected_aperture_type",
        [
            ("entrance_pupil_diameter", 3.0, "EntrancePupilDiameter"),
            ("float_by_stop_size", None, "FloatByStopSize"),
            ("image_f_number", 1.0, "ImageSpaceFNum"),
            ("object_numeric_aperture", 0.1, "ObjectSpaceNA"),
        ],
    )
    def test_aperture(self, aperture_type, aperture_value, expected_aperture_type, opticstudio_backend):
        opticstudio_backend.update_settings(aperture_type=aperture_type, aperture_value=aperture_value)

        assert opticstudio_backend.oss.SystemData.Aperture.ApertureType == zp.constants.process_constant(
            zp.constants.SystemData.ZemaxApertureType, expected_aperture_type
        )

        if aperture_type != "float_by_stop_size":
            assert opticstudio_backend.oss.SystemData.Aperture.ApertureValue == aperture_value

    @pytest.mark.parametrize(
        "ray_aiming,expected",
        [
            ("off", "Off"),
            ("paraxial", "Paraxial"),
            ("real", "Real"),
        ],
    )
    def test_ray_aiming(self, ray_aiming, expected, opticstudio_backend):
        opticstudio_backend.update_settings(ray_aiming=ray_aiming)

        assert opticstudio_backend.oss.SystemData.RayAiming.RayAiming == zp.constants.process_constant(
            zp.constants.SystemData.RayAimingMethod, expected
        )



from __future__ import annotations

import math
from types import MethodType, SimpleNamespace

import pytest
from optiland.surfaces import Surface

from tests.helpers import build_args
from visisipy.models import EyeModel
from visisipy.types import SampleSize


class MockSurface:
    def __init__(self):
        self.semi_aperture = 1.0
        self.changed_semi_aperture = False

    @property
    def semi_aperture(self):
        return self._semi_aperture

    @semi_aperture.setter
    def semi_aperture(self, value):
        self.changed_semi_aperture = True
        self._semi_aperture = value


class MockOptilandModel:
    def __init__(self):
        self._pupil = SimpleNamespace(surface=MockSurface())

    @property
    def pupil(self):
        return self._pupil


def patch_surface(surface: Surface, monkeypatch):
    def set_semi_aperture(self, a):
        self.changed_semi_aperture = True
        self.semi_aperture = a

    monkeypatch.setattr(surface, "set_semi_aperture", MethodType(set_semi_aperture, Surface))
    monkeypatch.setattr(surface, "changed_semi_aperture", False, raising=False)


class TestRefractionAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,sampling,pupil_diameter,field_type,use_higher_order_aberrations",
        [
            (None, None, None, None, None, None),
            ((0, 0), 0.543, None, None, None, None),
            ((1, 1), 0.632, 64, None, None, None),
            ((math.pi, math.tau), 0.543, "64x64", 0.5, None, None),
            ((0, 0), 0.543, "64x64", 0.5, "angle", None),
            pytest.param(
                (1, 1),
                0.632,
                SampleSize(32),
                0.5,
                "object_height",
                None,
            ),
            ((0, 0), 0.543, "32x32", 0.5, "angle", True),
            pytest.param(
                (1, 1),
                0.632,
                "64x64",
                0.5,
                "object_height",
                False,
            ),
        ],
    )
    def test_refraction(
        self,
        field_coordinate,
        wavelength,
        sampling,
        pupil_diameter,
        field_type,
        use_higher_order_aberrations,
        optiland_backend,
        optiland_analysis,
    ):
        optiland_backend.build_model(EyeModel(), object_distance=10 if field_type == "object_height" else float("inf"))

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            sampling=sampling,
            pupil_diameter=pupil_diameter,
            field_type=field_type,
            use_higher_order_aberrations=use_higher_order_aberrations,
            non_null_defaults={
                "sampling",
                "field_type",
                "use_higher_order_aberrations",
            },
        )

        assert optiland_analysis.refraction(**args)

    @pytest.mark.parametrize(
        "pupil_diameter,changed_pupil_diameter",
        [
            (None, False),
            (0.5, True),
        ],
    )
    def test_refraction_change_pupil(
        self,
        pupil_diameter,
        changed_pupil_diameter,
        optiland_backend,
        optiland_analysis,
    ):
        optiland_backend.build_model(EyeModel())

        # Manually set the aperture to a sentinel value
        new_semi_aperture = object()
        optiland_analysis.backend.model.pupil.surface.set_semi_aperture(new_semi_aperture)

        assert optiland_analysis.backend.model.pupil.surface.semi_aperture is new_semi_aperture

        optiland_analysis.refraction(pupil_diameter=pupil_diameter)

        if changed_pupil_diameter:
            assert optiland_analysis.backend.model.pupil.surface.semi_aperture == pupil_diameter / 2
        else:
            assert optiland_analysis.backend.model.pupil.surface.semi_aperture is new_semi_aperture



from __future__ import annotations

import numpy as np
import pytest
from optiland.materials import AbbeMaterial, IdealMaterial
from optiland.optic import Optic

from visisipy.models.materials import NavarroMaterials


def build_model(wavelength: float, n: float, abbe: float) -> Optic:
    """Simple optical system with 2 parallel surfaces: a stop / refracting interface, and image surface.
    All surfaces except the refracting interface and image surface have a refractive index of 1.0.
    The refractive index of the image surface is equal to that of the refracting interface, to prevent reflection.

    When a single ray trace is performed, the angle of incidence at the image surface is equal to the angle of
    refraction at the refracting interface. This allows to calculate the refractive index of the refracting interface
    from the angle of incidence at the refracting interface and the angle of incidence at the image surface.
    """
    system = Optic()

    system.set_field_type("angle")
    system.add_field(x=0, y=20)
    system.set_aperture("EPD", 1.0)
    system.add_wavelength(wavelength)

    air = IdealMaterial(n=1)
    test_material = AbbeMaterial(n=n, abbe=abbe)

    system.add_surface(index=0, comment="object", thickness=float("inf"), material=air)
    system.add_surface(index=1, comment="stop", is_stop=True, thickness=5.0, material=test_material)
    system.add_surface(index=2, comment="image", thickness=0.0, material=test_material)

    return system


@pytest.mark.parametrize(
    "material_model,wavelength,expected_index",
    [
        # Cornea
        (NavarroMaterials().cornea, 0.458, 1.3828),
        (NavarroMaterials().cornea, 0.543, 1.3777),
        (NavarroMaterials().cornea, 0.5893, 1.376),
        (NavarroMaterials().cornea, 0.6328, 1.3747),
        # Aqueous
        (NavarroMaterials().aqueous, 0.458, 1.3445),
        (NavarroMaterials().aqueous, 0.543, 1.3391),
        (NavarroMaterials().aqueous, 0.5893, 1.3374),
        (NavarroMaterials().aqueous, 0.6328, 1.336),
        # Lens
        (NavarroMaterials().lens, 0.458, 1.4292),
        (NavarroMaterials().lens, 0.543, 1.4222),
        (NavarroMaterials().lens, 0.5893, 1.42),
        (NavarroMaterials().lens, 0.6328, 1.4183),
        # Vitreous
        (NavarroMaterials().vitreous, 0.458, 1.3428),
        (NavarroMaterials().vitreous, 0.543, 1.3377),
        (NavarroMaterials().vitreous, 0.5893, 1.336),
        (NavarroMaterials().vitreous, 0.6328, 1.3347),
    ],
)
def test_material_model_refractive_index(material_model, wavelength, expected_index):
    system = build_model(wavelength, material_model.refractive_index, material_model.abbe_number)

    system.trace_generic(0, 1, 0, 0, wavelength=wavelength)
    trace_z, trace_y = system.surface_group.z, system.surface_group.y
    sin_angle_out = trace_y[2] / np.sqrt(trace_z[2] ** 2 + trace_y[2] ** 2)  # Assuming the ray passes through (0, 0)
    refractive_index = np.sin(np.deg2rad(20)) / sin_angle_out

    assert (trace_z[1], trace_y[1]) == (0, 0), "Ray must pass through the pupil center"
    assert refractive_index == pytest.approx(expected_index, rel=1e-4)



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from visisipy.models import EyeModel

pytestmark = [pytest.mark.needs_opticstudio]


class TestCardinalPointsAnalysis:
    @pytest.mark.parametrize(
        "surface_1,surface_2,expectation",
        [
            (None, None, does_not_raise()),
            (1, 6, does_not_raise()),
            (2, 4, does_not_raise()),
            (
                -1,
                7,
                pytest.raises(
                    ValueError,
                    match="surface_1 and surface_2 must be between 1 and the number of surfaces in the system",
                ),
            ),
            (
                1,
                8,
                pytest.raises(
                    ValueError,
                    match="surface_1 and surface_2 must be between 1 and the number of surfaces in the system",
                ),
            ),
            (3, 3, pytest.raises(ValueError, match="surface_1 must be less than surface_2")),
            (4, 2, pytest.raises(ValueError, match="surface_1 must be less than surface_2")),
        ],
    )
    def test_cardinal_points(self, opticstudio_backend, opticstudio_analysis, surface_1, surface_2, expectation):
        opticstudio_backend.build_model(EyeModel())

        with expectation:
            assert opticstudio_analysis.cardinal_points(surface_1, surface_2)



from __future__ import annotations

from typing import Any


def build_args(*, non_null_defaults: set[str], **kwargs) -> dict[str, Any]:
    """Build a dictionary of arguments for the analysis.

    Parameters
    ----------
    non_null_defaults : set[str], optional
        A list of argument names that are optional but do not have `None` as a default value.
    **kwargs : Any
        The arguments to be included in the dictionary.

    Returns
    -------
    dict[str, Any]
        A dictionary of arguments for the analysis.
    """
    null_defaults = kwargs.keys() - non_null_defaults

    result = {}

    for key in null_defaults:
        result[key] = kwargs[key]

    for key in non_null_defaults:
        if key in kwargs and kwargs[key] is not None:
            result[key] = kwargs[key]

    return result



from __future__ import annotations

import platform
from typing import Any

import pytest

import visisipy.optiland
from visisipy import backend
from visisipy.optiland.backend import OptilandBackend

if platform.system() == "Windows":
    import visisipy.opticstudio
    from visisipy.opticstudio.backend import OpticStudioBackend
else:
    OpticStudioBackend = object()

# ruff: noqa: SLF001
# pyright: reportOptionalMemberAccess=false, reportTypedDictNotRequiredAccess=false


class MockBackend:
    """Mock backend for testing purposes.

    Unlike the real backends, instances of this class are used instead of the class itself.
    This is necessary to make sure modifications to the backend settings are not shared between tests.
    """

    __name__ = "MockBackend"

    def __init__(self):
        self.settings: dict[str, Any] = {}
        self.initialized: bool = False
        self.oss = object()
        self.optic = object()

    def initialize(self, **settings):
        self.initialized = True
        self.settings = settings

    def update_settings(self, **settings):
        if len(settings) > 0:
            self.settings.update(settings)


@pytest.fixture
def mock_backend(monkeypatch):
    monkeypatch.setattr(backend, "_BACKEND", MockBackend())


@pytest.fixture
def mock_opticstudio_backend(monkeypatch):
    instance = MockBackend()
    monkeypatch.setattr(visisipy.opticstudio, "OpticStudioBackend", instance)

    return instance


@pytest.fixture
def mock_optiland_backend(monkeypatch):
    instance = MockBackend()
    monkeypatch.setattr(visisipy.optiland, "OptilandBackend", instance)

    return instance


class TestSetBackend:
    @pytest.mark.windows_only
    @pytest.mark.parametrize("name", ["opticstudio", backend.Backend.OPTICSTUDIO])
    def test_set_opticstudio_backend(self, name, mock_opticstudio_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", None)  # Reset the backend to avoid side effects in other tests

        backend.set_backend(name, field_type="object_height")

        assert mock_opticstudio_backend == backend._BACKEND
        assert mock_opticstudio_backend.initialized
        assert mock_opticstudio_backend.settings["field_type"] == "object_height"

    @pytest.mark.parametrize("name", ["optiland", backend.Backend.OPTILAND])
    def test_set_optiland_backend(self, name, mock_optiland_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", None)  # Reset the backend to avoid side effects in other tests

        backend.set_backend(name, aperture_type="image_f_number")

        assert mock_optiland_backend == backend._BACKEND
        assert mock_optiland_backend.initialized
        assert mock_optiland_backend.settings["aperture_type"] == "image_f_number"

    def test_set_backend_twice_raises_warning(self, mock_optiland_backend):
        backend.set_backend("optiland")
        with pytest.warns(UserWarning, match="The backend is already set to MockBackend"):
            backend.set_backend("optiland")

    @pytest.mark.filterwarnings("ignore:The backend is already set to MockBackend")
    def test_set_unknown_backend(self):
        with pytest.raises(ValueError, match="Unknown backend: unknown"):
            backend.set_backend("unknown")


class TestGetBackend:
    @pytest.mark.windows_only
    def test_get_backend(self, mock_opticstudio_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", mock_opticstudio_backend)

        assert backend.get_backend() == mock_opticstudio_backend

    def test_get_backend_sets_backend(self, mock_optiland_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", None)  # Reset the backend so get_backend() sets it
        monkeypatch.setattr(backend, "_DEFAULT_BACKEND", backend.Backend.OPTILAND)

        assert backend._BACKEND is None

        assert backend.get_backend() == mock_optiland_backend


class TestGetModels:
    @pytest.mark.windows_only
    def test_get_oss(self, mock_opticstudio_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", mock_opticstudio_backend)

        assert backend.get_oss() is mock_opticstudio_backend.oss

    @pytest.mark.windows_only
    def test_get_oss_no_opticstudio_backend(self, mock_optiland_backend):
        assert backend.get_oss() is None

    def test_get_optic(self, mock_optiland_backend, monkeypatch):
        monkeypatch.setattr(backend, "_BACKEND", mock_optiland_backend)

        assert backend.get_optic() is mock_optiland_backend.optic

    def test_get_optic_no_optiland_backend(self, mock_backend):
        assert backend.get_optic() is None


class TestUpdateSettings:
    def test_update_settings(self, mock_backend):
        field_type = "angle"
        fields = [(0, 0), (1, 1)]
        wavelengths = [550]

        backend.update_settings(field_type=field_type, fields=fields, wavelengths=wavelengths)

        assert backend._BACKEND.settings["field_type"] == "angle"
        assert backend._BACKEND.settings["fields"] == [(0, 0), (1, 1)]
        assert backend._BACKEND.settings["wavelengths"] == [550]

    @pytest.mark.filterwarnings("ignore:The OpticStudio backend settings")
    @pytest.mark.parametrize(
        "backend_type", [pytest.param(OpticStudioBackend, marks=pytest.mark.windows_only), OptilandBackend]
    )
    def test_update_settings_by_backend(self, backend_type):
        field_type = "object_height"
        fields = [(0, 0), (1, 1)]
        wavelengths = [550]

        backend.update_settings(backend_type, field_type=field_type, fields=fields, wavelengths=wavelengths)

        assert backend_type.settings["field_type"] == "object_height"
        assert backend_type.settings["fields"] == [(0, 0), (1, 1)]
        assert backend_type.settings["wavelengths"] == [550]


class TestGetSetting:
    def test_get_setting(self, monkeypatch):
        monkeypatch.setattr(backend.BaseBackend, "settings", {"test_setting": "value"}, raising=False)

        assert backend.BaseBackend.get_setting("test_setting") == "value"

    def test_get_undefined_setting_raises_keyerror(self, monkeypatch):
        monkeypatch.setattr(backend.BaseBackend, "settings", {}, raising=False)

        with pytest.raises(KeyError, match="Setting 'undefined_setting' does not exist"):
            backend.BaseBackend.get_setting("undefined_setting")



"""Calculate the cardinal points of an eye model."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, overload

from visisipy.analysis.base import _AUTOMATIC_BACKEND, analysis

if TYPE_CHECKING:
    from visisipy import EyeModel
    from visisipy.backend import BaseBackend

__all__ = ("CardinalPoints", "CardinalPointsResult", "cardinal_points")


class CardinalPoints(NamedTuple):
    """The cardinal points of a system in object and image space."""

    object: float
    image: float


@dataclass
class CardinalPointsResult:
    """The cardinal points of a system.

    Attributes
    ----------
    focal_lengths : CardinalPoints
        The focal lengths of the system.
    focal_points : CardinalPoints
        The focal points of the system.
    principal_points : CardinalPoints
        The principal points of the system.
    anti_principal_points : CardinalPoints
        The anti-principal points of the system.
    nodal_points : CardinalPoints
        The nodal points of the system.
    anti_nodal_points : CardinalPoints
        The anti-nodal points of the system.
    """

    focal_lengths: CardinalPoints
    focal_points: CardinalPoints
    principal_points: CardinalPoints
    anti_principal_points: CardinalPoints
    nodal_points: CardinalPoints
    anti_nodal_points: CardinalPoints


@overload
def cardinal_points(
    model: EyeModel | None = None,
    surface_1: int | None = None,
    surface_2: int | None = None,
    *,
    return_raw_result: Literal[False] = False,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> CardinalPointsResult: ...


@overload
def cardinal_points(
    model: EyeModel | None = None,
    surface_1: int | None = None,
    surface_2: int | None = None,
    *,
    return_raw_result: Literal[True] = True,
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[CardinalPointsResult, Any]: ...


@analysis
def cardinal_points(
    model: EyeModel | None = None,  # noqa: ARG001
    surface_1: int | None = None,
    surface_2: int | None = None,
    *,
    return_raw_result: bool = False,  # noqa: ARG001
    backend: type[BaseBackend] = _AUTOMATIC_BACKEND,
) -> tuple[CardinalPointsResult, Any]:
    """Get the cardinal points of the system between `surface_1` and `surface_2`.

    Parameters
    ----------
    model : EyeModel | None
        The eye model to be used in the ray trace. If `None`, the current eye model will be used.
    surface_1 : int | None, optional
        The first surface to be used in the analysis. If `None`, the first surface in the system will be used.
        Defaults to `None`.
    surface_2 : int | None, optional
        The second surface to be used in the analysis. If `None`, the last surface in the system will be used.
        Defaults to `None`.
    return_raw_result : bool, optional
        Return the raw analysis result from the backend. Defaults to `False`.
    backend : type[BaseBackend]
        The backend to be used for the analysis. If not provided, the default backend is used.

    Returns
    -------
    CardinalPointsResult
        The cardinal points of the system.
    Any
        The raw analysis result from the backend.
    """
    return backend.analysis.cardinal_points(
        surface_1=surface_1,
        surface_2=surface_2,
    )



from __future__ import annotations

import pytest

from tests.helpers import build_args
from visisipy.models import EyeModel
from visisipy.types import SampleSize

pytestmark = [pytest.mark.needs_opticstudio]


class TestZernikeStandardCoefficientsAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,field_type,sampling,maximum_term",
        [
            (None, None, None, None, None),
            ((0, 0), 0.543, None, None, None),
            ((1, 1), 0.632, "angle", None, None),
            ((0.5, 0.5), 0.543, "object_height", 64, None),
            ((0, 0), 0.543, "angle", SampleSize(64), 45),
            ((1, 1), 0.632, "angle", "64x64", 100),
        ],
    )
    def test_zernike_standard_coefficients(
        self,
        field_coordinate,
        wavelength,
        field_type,
        sampling,
        maximum_term,
        opticstudio_backend,
        opticstudio_analysis,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            field_type=field_type,
            sampling=sampling,
            maximum_term=maximum_term,
            non_null_defaults={"field_type", "sampling", "maximum_term"},
        )

        assert opticstudio_analysis.zernike_standard_coefficients(**args)

    def test_zernike_standard_coefficients_maximum_term(self, opticstudio_analysis):
        coefficients, _ = opticstudio_analysis.zernike_standard_coefficients(maximum_term=20)

        assert coefficients[21] == 0



from __future__ import annotations

from contextlib import nullcontext as does_not_raise

import pytest

from tests.helpers import build_args
from visisipy.models import EyeModel


class TestRayTraceAnalysis:
    @pytest.mark.parametrize(
        "coordinates,wavelengths,field_type,pupil,expectation",
        [
            (None, None, None, None, does_not_raise()),
            ([(0, 0), (1, 1)], None, None, None, does_not_raise()),
            ([(0, 0), (1, 1)], [0.543], None, None, does_not_raise()),
            ([(0, 0), (1, 1)], [0.543], "angle", None, does_not_raise()),
            pytest.param(
                [(0, 0), (1, 1)],
                [0.543, 0.600],
                "object_height",
                (1, 0),
                does_not_raise(),
            ),
            (
                [(0, 0), (1, 1)],
                [0.543],
                "angle",
                (2, 0),
                pytest.raises(ValueError, match="Pupil coordinates must be between -1 and 1"),
            ),
            (
                [(0, 0), (1, 1)],
                [0.543, 0.600],
                "angle",
                (0, 2),
                pytest.raises(ValueError, match="Pupil coordinates must be between -1 and 1"),
            ),
        ],
    )
    def test_raytrace(
        self,
        coordinates,
        wavelengths,
        field_type,
        pupil,
        expectation,
        optiland_backend,
        optiland_analysis,
    ):
        optiland_backend.build_model(EyeModel(), object_distance=10 if field_type == "object_height" else float("inf"))

        args = build_args(
            non_null_defaults={"field_type", "pupil"},
            coordinates=coordinates,
            wavelengths=wavelengths,
            field_type=field_type,
            pupil=pupil,
        )

        with expectation:
            assert optiland_analysis.raytrace(**args)

    def test_raytrace_dataframe_structure(self, optiland_backend, optiland_analysis):
        optiland_backend.build_model(EyeModel())

        coordinates = [(0, 0), (1, 1)]

        result, _ = optiland_analysis.raytrace(coordinates)
        expected_columns = {
            "index",
            "x",
            "y",
            "z",
            "field",
            "wavelength",
            "surface",
            "comment",
        }

        assert set(result.columns) == expected_columns
        assert set(result.field.unique()) == set(coordinates)



from __future__ import annotations

import platform
from types import ModuleType

import pytest

# ruff: noqa: PLC0415


@pytest.mark.windows_only
class TestZospyImportWindows:
    def test_zospy_import(self):
        import zospy as zp

        assert isinstance(zp, ModuleType)

    def test_opticstudio_in_all(self):
        import visisipy

        assert "opticstudio" in visisipy.__all__


@pytest.mark.skipif(platform.system() == "Windows", reason="No ImportError raised on Windows")
class TestZospyImportNonWindows:
    @pytest.mark.parametrize(
        "import_statement",
        ["import zospy", "import zospy as zp", "from zospy import analyses", "import zospy.constants as constants"],
    )
    def test_zospy_import_raises_importerror(self, import_statement):
        with pytest.raises(ImportError, match="Could not import module 'zospy'"):
            exec(import_statement)

    def test_opticstudio_not_in_all(self):
        import visisipy

        assert "opticstudio" not in visisipy.__all__
